@@ -1,118 +1,21 @@
-#!/usr/local/perl-5.00404/bin/perl -Tw
+#!/usr/bin/perl -wT
 #
-# $Id: guestbook.pl,v 1.28 2002/02/14 12:57:53 nickjc Exp $
-#
-# $Log: guestbook.pl,v $
-# Revision 1.28  2002/02/14 12:57:53  nickjc
-# * file locking review
-#
-# Revision 1.27  2002/01/29 08:50:44  nickjc
-# Improved check_url_valid function from FormMail.pl
-#
-# Revision 1.26  2001/12/28 22:17:07  nickjc
-# minor HTML filter fix
-#
-# Revision 1.25  2001/12/21 08:53:24  nickjc
-# More HTML filter fixes
-#
-# Revision 1.24  2001/12/20 08:57:54  nickjc
-# tidying and adding comments
-#
-# Revision 1.23  2001/12/19 23:12:23  nickjc
-# HTML filter fixes
-#
-# Revision 1.22  2001/12/18 22:21:13  nickjc
-# * minor HTML filter fixes
-# * started on allowing the style attribute
-# * added non-XHTML (but harmless) NOBR tag
-#
-# Revision 1.21  2001/12/17 23:01:52  nickjc
-# HTML filter fixes
-#
-# Revision 1.20  2001/12/16 11:46:14  nickjc
-# Ensure regexp matches all text in cleanup_html()
-#
-# Revision 1.19  2001/12/15 22:22:59  nickjc
-# * added $url validity check
-# * misc minor fixes and tidying
-#
-# Revision 1.18  2001/12/15 00:19:25  nickjc
-# Added a whitelist-based HTML filter to strip out unsafe constructs
-# when $allow_html is 1
-#
-# Revision 1.17  2001/12/11 08:52:16  nickjc
-# * more general unescape_html(), in preparation for proper HTML filter
-# * more aggressive escape_html()
-# * don't turn entities like '&lt;' into '&amp;lt;' in strip_html() if
-#   $allow_html is 0
-#
-# Revision 1.16  2001/12/10 23:34:37  nickjc
-# Strengthened strip_html() when $allow_html is false
-#
-# Revision 1.15  2001/12/09 23:33:53  nickjc
-# bug fix: strip_html was breaking email sending by adding SSI exploit
-# defence stuff to the email address.
-#
-# Revision 1.14  2001/12/01 17:53:11  gellyfish
-# * Fixed up some 5.004.04 compatibility issues
-# * Started rationalization of HTML handling
-#
-# Revision 1.13  2001/12/01 11:44:43  gellyfish
-# SSI exploit fix as suggested by Pete Sargent
-#
-# Revision 1.12  2001/11/26 13:40:05  nickjc
-# Added \Q \E around variables in regexps where metacharacters in the
-# variables shouldn't be interpreted by the regex engine.
-#
-# Revision 1.11  2001/11/25 15:28:43  gellyfish
-# * Added security features
-# * more refactoring
-#
-# Revision 1.10  2001/11/25 11:39:38  gellyfish
-# * add missing use vars qw($DEBUGGING) from most of the files
-# * sundry other compilation failures
-#
-# Revision 1.9  2001/11/24 11:59:58  gellyfish
-# * documented strfime date formats is various places
-# * added more %ENV cleanup
-# * spread more XHTML goodness and CSS stylesheet
-# * generalization in wwwadmin.pl
-# * sundry tinkering
-#
-# Revision 1.8  2001/11/19 09:21:44  gellyfish
-# * added allow_html functionality
-# * fixed potential for pre lock clobbering in guestbook
-# * some XHTML toshing
-#
-# Revision 1.7  2001/11/16 09:06:53  gellyfish
-# Had forgotten to declare $DEBUGGING
-#
-# Revision 1.6  2001/11/14 23:00:13  davorg
-# More XHTML fixes.
-#
-# Revision 1.5  2001/11/14 22:21:17  davorg
-# Changed script archive URL to Sourceforge.
-# Added link to support mailing list.
-#
-# Revision 1.4  2001/11/14 19:39:14  davorg
-# Fixed stupid bug getting CGI parameters
-#
-# Revision 1.3  2001/11/13 20:35:14  gellyfish
-# Added the CGI::Carp workaround
-#
-# Revision 1.2  2001/11/11 17:55:27  davorg
-# Small amount of post-import tidying :)
-#
-# Revision 1.1.1.1  2001/11/11 16:48:53  davorg
-# Initial import
+# $Id: guestbook.pl,v 1.37 2002/04/15 07:46:30 nickjc Exp $
 #
 
 use strict;
 use POSIX qw(strftime);
 use CGI qw(:standard);
 use Fcntl qw(:DEFAULT :flock);
+use IO::File;
 
-use vars qw($DEBUGGING @debug_msg);
+use vars qw(
+  $DEBUGGING $done_headers @debug_msg $guestbookurl
+  $guestbookreal $guestlog $cgiurl $emulate_matts_code
+  $style $mail $uselog $linkmail $separator $redirection
+  $entry_order $remote_mail $allow_html $line_breaks
+  $mailprog $recipient $short_date_fmt $long_date_fmt
+);
 
 # sanitize the environment
 
@@ -132,10 +35,10 @@ BEGIN
    $DEBUGGING = 1;
 }
 
-my $guestbookurl  = 'http://your.host.com/~yourname/guestbook.html';
-my $guestbookreal = '/home/yourname/public_html/guestbook.html';
-my $guestlog      = '/home/yourname/public_html/guestlog.html';
-my $cgiurl        = 'http://your.host.com/cgi-bin/guestbook.pl';
+$guestbookurl  = 'http://your.host.com/~yourname/guestbook.html';
+$guestbookreal = '/home/yourname/public_html/guestbook.html';
+$guestlog      = '/home/yourname/public_html/guestlog.html';
+$cgiurl        = 'http://your.host.com/cgi-bin/guestbook.pl';
 
 # $emulate_matts_code determines whether the program should behave exactly
 # like the original guestbook program.  It should be set to 1 if you
@@ -144,37 +47,37 @@ my $cgiurl        = 'http://your.host.com/cgi-bin/guestbook.pl';
 # then potentially it will not work with files produced by the original
 # version - this is recommended for people installing this for the first time.
 
-my $emulate_matts_code = 1;
+$emulate_matts_code = 1;
 
 # $style is the URL of a CSS stylesheet which will be used for script
 # generated messages.  This probably want's to be the same as the one
 # that you use for all the other pages.  This should be a local absolute
 # URI fragment.
 
-my $style = '/css/nms.css';
+$style = '/css/nms.css';
 
 
-my $mail        = 0;
-my $uselog      = 1;
-my $linkmail    = 1;
-my $separator   = 1;
-my $redirection = 0;
-my $entry_order = 1;
-my $remote_mail = 0;
-my $allow_html  = 0;
-my $line_breaks = 0;
+$mail        = 0;
+$uselog      = 1;
+$linkmail    = 1;
+$separator   = 1;
+$redirection = 0;
+$entry_order = 1;
+$remote_mail = 0;
+$allow_html  = 0;
+$line_breaks = 0;
 
 # $mailprog is the program that will be used to send mail if that is
 # required.  It should be the full path of a program that will accept
 # the message on its standard input, it should also include any required
 # switches.  If $mail is set to 0 above this can be ignores.
 
-my $mailprog  = '/usr/lib/sendmail -t -oi -oem';
+$mailprog  = '/usr/lib/sendmail -t -oi -oem';
 
 # $recipient is the address of the person who should be mailed if $mail is
 # set to 1 above.
 
-my $recipient = 'you@your.com';
+$recipient = 'you@your.com';
 
 # $long_date_fmt and $short_date_fmt describe the format of the dates that
 # will output - the replacement parameters you can use here are:
@@ -192,8 +95,8 @@ my $recipient = 'you@your.com';
 # %T - the time in 24 hour format (%H:%M:%S)
 # %Z - the time zone (full name or abbreviation)
 
-my $long_date_fmt  = '%A, %B %d, %Y at %T (%Z)';
-my $short_date_fmt = '%d/%m/%y %T %Z';
+$long_date_fmt  = '%A, %B %d, %Y at %T (%Z)';
+$short_date_fmt = '%d/%m/%y %T %Z';
 
 # End configuration
 
@@ -208,7 +111,7 @@ BEGIN
    {
       my ( $message ) = @_;
 
-      if ( $main::DEBUGGING )
+      if ( $DEBUGGING )
       {
          $message =~ s/</&lt;/g;
          $message =~ s/>/&gt;/g;
@@ -223,10 +126,13 @@ BEGIN
 
       return undef if $file =~ /^\(eval/;
 
-      print "Content-Type: text/html\n\n";
+      print "Content-Type: text/html; charset=iso-8859-1\n\n" unless $done_headers;
 
       print <<EOERR;
-<html>
+<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
+    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
+
+<html xmlns="http://www.w3.org/1999/xhtml">
   <head>
     <title>Error</title>
   </head>
@@ -247,141 +153,107 @@ EOERR
    $SIG{__DIE__} = \&fatalsToBrowser;
 }
 
-my @now       = localtime();
-my $date      = strftime($long_date_fmt, @now);
-my $shortdate = strftime($short_date_fmt, @now);
+use vars qw($style_element);
+$style_element = $style ?
+                 qq%<link rel="stylesheet" type="text/css" href="$style" />%
+               : '';
 
-my ($username, $realname, $comments)
-  = (param('username'), param('realname'), param('comments'));
+use vars qw($date $shortdate);
+my @now    = localtime();
+$date      = strftime($long_date_fmt, @now);
+$shortdate = strftime($short_date_fmt, @now);
 
-my ($city, $state, $country)
-  = (param('city'), param('state'), param('country'));
-
-my ($url) = param('url');
-$url = '' if $url and not check_url_valid($url);
+my @input_names =
+  qw(username realname comments city state country url);
+use vars qw(%inputs);
+foreach my $input (@input_names) {
+  $inputs{$input} = strip_nonprintable(param($input));
+}
 
 # There is a possibility that the comments can be escaped if passed as
 # the hidden field from the form_error() form
+if (param('encoded_comments')) {
+  $inputs{comments} = unescape_html( $inputs{comments} );
+}
 
-my $encoded_comments = param('encoded_comments') || 0;
+$inputs{'url'}    = '' unless check_url_valid($inputs{'url'});
+$inputs{username} = '' unless check_email($inputs{username});
 
-form_error('no_comments') unless $comments;
+# Strip out HTML unless we are allowing it.  The process_html
+# sub should take care of everything.
+use vars qw($comments);
+$comments = process_html($inputs{comments}, $line_breaks, $allow_html);
 
-$comments = unescape_html($comments) if $encoded_comments;
+# Generate versions of the inputs with HTML metacharacters
+# escaped - HTML should not # be allowed anywhere but the
+# comment.
+use vars qw(%escaped);
+%escaped = map {$_ => escape_html($inputs{$_})} keys %inputs;
 
-# Strip out HTML unless we are allowing it
-# process_html should take care of everything.
+form_error('no_comments') unless $inputs{comments};
+form_error('no_name')     unless $inputs{realname};
 
-$comments = process_html($comments, $line_breaks, $allow_html);
+rewrite_file($guestbookreal, sub
+{
+   if (defined and /<!--begin-->/) {
 
-form_error('no_name')     unless $realname;
+     $_ = '' unless $entry_order;
 
+     $_ .= "<b>$comments</b><br />\n";
 
-# Get rid of the $username unless it is a valid e-mail address
-
-$username = '' unless check_email($username);
-
-# Escape any HTML in the rest of the fields - HTML should not
-# be allowed anywhere but the comment.
-
-my %escaped = (
- username => escape_html($username),
- realname => escape_html($realname),
- city     => escape_html($city),
- state    => escape_html($state),
- country  => escape_html($country),
-);
-
-open (LOCK, ">>$guestbookreal.lck")
-  || die "Can't Open $guestbookreal.lck: $!\n";
-flock LOCK, LOCK_EX
-  || die "Can't lock $guestbookreal.lck: $!\n";
-
-open (GUEST_IN, "<$guestbookreal")
-  || die "Can't Open $guestbookreal: $!\n";
-open (GUEST, ">$guestbookreal.tmp")
-  || die "Can't Open $guestbookreal.tmp: $!\n";
-
-
-while ( defined( $_ = <GUEST_IN> ) ) {
-   if (/<!--begin-->/) {
-
-     if ($entry_order) {
-       print GUEST "<!--begin-->\n";
+     if ($inputs{'url'}) {
+       $_ .= qq(<a href="$escaped{'url'}">$escaped{realname}</a>);
+     } else {
+       $_ .= $escaped{realname};
      }
 
-
-     print GUEST "<b>$comments</b><br />\n";
-
-     if ($url) {
-       my $eurl = escape_html($url);
-       print GUEST qq(<a href="$eurl">$escaped{realname}</a>);
-      } else {
-         print GUEST $escaped{realname};
-      }
-
-     if ($username){
+     if ($inputs{username}){
        if ($linkmail) {
-         print GUEST qq( &lt;<a href="mailto:$escaped{username}">);
-         print GUEST "$escaped{username}</a>&gt;";
+         $_ .= qq( &lt;<a href="mailto:$escaped{username}">);
+         $_ .= "$escaped{username}</a>&gt;";
        } else {
-         print GUEST " &lt;$escaped{username}&gt;";
+         $_ .= " &lt;$escaped{username}&gt;";
        }
      }
 
-     print GUEST "<br />\n";
+     $_ .= "<br />\n";
 
-     if ($city){
-       print GUEST "$escaped{city}, ";
+     if ($inputs{city}){
+       $_ .= "$escaped{city}, ";
      }
 
-     if ($state){
-       print GUEST $escaped{state};
+     if ($inputs{state}){
+       $_ .= $escaped{state};
      }
 
-     if ($country){
-       print GUEST " $escaped{country}";
+     if ($inputs{country}){
+       $_ .= " $escaped{country}";
      }
 
      if ($separator) {
-       print GUEST " - $date<hr />\n\n";
+       $_ .= " - $date<hr />\n\n";
      } else {
-       print GUEST " - $date<p />\n\n";
+       $_ .= " - $date<p />\n\n";
      }
 
-     unless ($entry_order) {
-       print GUEST "<!--begin-->\n";
-     }
-
-   } else {
-     print GUEST $_;
+     $_ .= "<!--begin-->\n" unless $entry_order;
    }
-}
-
-unless ( close GUEST ) {
-   unlink "$guestbookreal.tmp";
-   die "write to $guestbookreal.tmp: $!";
-}
-
-rename "$guestbookreal.tmp", $guestbookreal
-   or die "rename $guestbookreal.tmp -> $guestbookreal: $!";
+});
 
 write_log('entry') if $uselog;
 
-close LOCK;
-
 if ($mail) {
    my $to = $recipient;
-   my $reply = "$username ($realname)";
-   my $from   = "$username ($realname)";
+   my $reply = "$inputs{username} ($inputs{realname})";
+   my $from   = "$inputs{username} ($inputs{realname})";
    my $subject = 'Entry to Guestbook';
    my $body    = 'You have a new entry in your guestbook:';
    do_mail($to, $from, $reply, $subject, $body );
 }
 
-if ($remote_mail && $username) {
+if ($remote_mail && $inputs{username}) {
 
-  my $to = $username;
+  my $to = $inputs{username};
   my $from = $recipient;
   my $reply = $recipient;
   my $subject = 'Entry to Guestbook';
@@ -400,12 +272,11 @@ sub form_error {
 
   my ( $why ) = @_;
 
-  $comments = escape_html($comments) if $comments;
-
   my ( $title, $heading, $text, $comments_field ) ;
 
   if ( $why eq 'no_name' ) {
-      $realname = '';
+      $inputs{realname} = '';
+      $escaped{realname} = '';
       $title = 'No Name';
       $heading = 'Your Name appears to be blank';
       $text =<<EOTEXT;
@@ -415,7 +286,7 @@ added.  Please add your name in the blank below.
 EOTEXT
       $comments_field =<<EOCOMMENT;
     Comments have been retained.
-        <input type="hidden" name="comments" value="$comments" />
+        <input type="hidden" name="comments" value="$escaped{comments}" />
         <input type="hidden" name="comments_encoded" value="1" />
 EOCOMMENT
    }
@@ -438,22 +309,23 @@ EOCOMMENT
       $text    = 'Please check your input and resubmit';
       $comments_field =<<EOCOMMENT;
       Comments:<br />
-      <textarea name="comments" cols="60" rows="4">$comments</textarea />
+      <textarea name="comments" cols="60" rows="4">$escaped{comments}</textarea />
       <input type="hidden" name="comments_encoded" value="1" />
 EOCOMMENT
    }
 
   local $^W; # suppress warnings as we may have missing fields;
 
-  print header;
+  print "Content-Type: text/html; charset=iso-8859-1\n\n";
+  $done_headers++;
   print <<END_FORM;
-<?xml version="1.0" encoding="UTF-8"?>
+<?xml version="1.0" encoding="iso-8859-1"?>
 <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
   "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
-<html>
+<html xmlns="http://www.w3.org/1999/xhtml">
   <head>
     <title>$title</title>
-     <link rel="stylesheet" type="text/css" href="$style" />
+    $style_element
   </head>
   <body>
     <h1>$heading</h1>
@@ -462,22 +334,22 @@ EOCOMMENT
     </p>
     <form method="post" action="$cgiurl">
       <p>Your Name: <input type="text" name="realname"
-                           value="$realname" size="30" /><br />
+                           value="$escaped{realname}" size="30" /><br />
         E-Mail: <input type="text" name="username"
-                       value="$username" size="40" /><br />
-        City: <input type="text" name="city" value="$city"
+                       value="$escaped{username}" size="40" /><br />
+        City: <input type="text" name="city" value="$escaped{city}"
                      size="15" />,
         State: <input type="text" name="state"
-                      value="$state" size="2" />
-        Country: <input type="text" name="country" value="$country"
+                      value="$escaped{state}" size="2" />
+        Country: <input type="text" name="country" value="$escaped{country}"
                         size="15" /></p>
       <p>
        $comments_field
       </p>
-      <p><input type="submit"> * <input type="reset"></p>
+      <p><input type="submit" /> * <input type="reset" /></p>
     </form>
     <hr />
-    <p>Return to the <a href="$guestbookurl">Guestbook</a>.
+    <p>Return to the <a href="$guestbookurl">Guestbook</a></p>.
   </body>
 </html>
 
@@ -494,47 +366,48 @@ END_FORM
 sub write_log {
   my ($log_type) = @_;
 
-  if ( open(LOG, ">>$guestlog") )
+  my $found_close_body = 0;
+
+  rewrite_file($guestlog, sub
   {
-      if ( flock LOG, LOCK_EX )
-      {
-
-         my $remote = remote_host();
-         if ($log_type eq 'entry') {
-           print LOG "$remote - [$shortdate]<br />\n";
-         } elsif ($log_type eq 'no_name') {
-           print LOG "$remote - [$shortdate] - ERR: No Name<br />\n";
-         } elsif ($log_type eq 'no_comments') {
-           print LOG "$remote - [$shortdate] - ERR: No Comments<br />\n";
-         }
-      }
-      else
-      {
-         die "Can't lock log file: $!\n" if $main::DEBUGGING;
-      }
-  }
-  else
-  {
-    # We probably dont wan't to show this to the crowd :)
-
-    die "Can't open log file: $!\n" if $DEBUGGING;
-
-  }
+     if (not defined)
+     {
+        # Matt's original guestlog.html is missing these close
+        # tags, so if we don't find </body> we append them to
+        # make guestlog.html into valid XHTML.
+        $_ = "</body>\n</html>\n" unless $found_close_body;
+     }
+     if (defined and m#</body>#i)
+     {
+        $found_close_body = 1;
+        my $remote = remote_host();
+        my $logline;
+        if ($log_type eq 'entry') {
+          $logline = "$remote - [$shortdate]<br />\n";
+        } elsif ($log_type eq 'no_name') {
+          $logline = "$remote - [$shortdate] - ERR: No Name<br />\n";
+        } elsif ($log_type eq 'no_comments') {
+          $logline = "$remote - [$shortdate] - ERR: No Comments<br />\n";
+        }
+        $_ = "$logline$_";
+     }
+  });
 
 }
 
 # Redirection Option
 sub no_redirection {
 
-  print header();
+  print "Content-Type: text/html; charset=iso-8859-1\n\n";
+  $done_headers++;
   print <<END_HTML;
-<?xml version="1.0" encoding="UTF-8"?>
+<?xml version="1.0" encoding="iso-8859-1"?>
 <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
   "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
-<html>
+<html xmlns="http://www.w3.org/1999/xhtml">
   <head>
     <title>Thank You</title>
-     <link rel="stylesheet" type="text/css" href="$style" />
+    $style_element
   </head>
   <body>
     <h1>Thank You For Signing The Guestbook</h1>
@@ -543,18 +416,17 @@ sub no_redirection {
       been added to the guestbook.</p>
     <hr />
     <p>Here is what you submitted:</p>
-    <p><b>$comments</b><br>
+    <p><b>$comments</b></p><br />
 
 END_HTML
 
-  if ($url) {
-    my $eurl = escape_html($url);
-    print qq(<a href="$eurl">$escaped{realname}</a>);
+  if ($inputs{'url'}) {
+    print qq(<a href="$escaped{'url'}">$escaped{realname}</a>);
   } else {
     print $escaped{realname};
   }
 
-  if ($username){
+  if ($inputs{username}){
     if ($linkmail) {
       print qq( &lt;<a href="mailto:$escaped{username}">);
       print "$escaped{username}</a>&gt;";
@@ -565,13 +437,13 @@ END_HTML
 
   print "<br />\n";
 
-  print "$escaped{city}," if $city;
+  print "$escaped{city}," if $inputs{city};
 
-  print " $escaped{state}" if $state;
+  print " $escaped{state}" if $inputs{state};
 
-  print " $escaped{country}" if $country;
+  print " $escaped{country}" if $inputs{country};
 
-  print " - $date<p>\n";
+  print " - $date\n";
 
   if (scalar @debug_msg) {
     print qq|<br /><font color="red">\n|;
@@ -610,20 +482,20 @@ Subject: $subject
 $body
 ------------------------------------------------------
 $comments
-$realname
+$inputs{realname}
 EOMAIL
 
-  if ($username){
-    print MAIL " <$username>";
+  if ($inputs{username}){
+    print MAIL " <$inputs{username}>";
   }
 
   print MAIL "\n";
 
-  print MAIL "$city'," if $city;
+  print MAIL "$inputs{city}'," if $inputs{city};
 
-  print MAIL " $state" if $state;
+  print MAIL " $inputs{state}" if $inputs{state};
 
-  print MAIL " $country" if $country;
+  print MAIL " $inputs{country}" if $inputs{country};
 
   print MAIL " - $date\n";
   print MAIL "------------------------------------------------------\n";
@@ -631,6 +503,13 @@ EOMAIL
   close (MAIL);
 }
 
+sub strip_nonprintable {
+  my $text = shift;
+  return '' unless defined $text;
+  $text=~ tr#\011\012\040-\176\240-\377##dc;
+  return $text;
+}
+	
 ##############################################################
 #
 # Validity checks for various contexts.
@@ -667,6 +546,82 @@ sub check_email {
   }
 }
 
+=head1 FILE MANIPULATION FUNCTIONS
+
+=over
+
+=item rewrite_file( FILENAME, CALLBACK )
+
+This function makes an atomic chanage to a file, by copying
+an old version to a new version line by line and then
+renaming the new version over the old version.  An external
+lock file is used to prevent clashes between several
+processes accessing the file at once.
+
+Dies on error.
+
+FILENAME is the filesystem path to the file.
+
+CALLBACK is a coderef to act on the contents of the file
+line by line.  It gets called once for each line in the
+file, with the line stored in C<$_>.  Any changes made to
+C<$_> will be reflected in the new version of the file.
+
+The CALLBACK coderef will be called one more time after
+all the lines have been processed, with C<$_> set to
+undef.
+
+=cut
+
+sub rewrite_file
+{
+   my ($filename, $callback) = @_;
+   local $_;
+
+   my $lock = IO::File->new(">>$filename.lck") or die
+      "open $filename.lck: $!";
+   flock $lock, LOCK_EX or die "flock $filename: $!";
+
+   my $temp = IO::File->new(">$filename.tmp") or die
+      "open >$filename.tmp: $!";
+
+   my $in = IO::File->new("<$filename") or die
+      "open <$filename: $!";
+
+   my $last_line_done = 0;
+   until ($last_line_done)
+   {
+      $last_line_done = not defined ($_ = <$in>);
+
+      &{ $callback }();
+      if (defined and length and not $temp->print($_))
+      {
+         my $write_err = $!;
+         $temp->close;
+         unlink "$filename.tmp";
+         die "write to $filename.tmp: $write_err";
+      }
+   }
+
+   unless ($temp->close)
+   {
+      my $close_err = $!;
+      unlink "$filename.tmp";
+      die "close $filename.tmp: $close_err";
+   }
+
+   $in->close;
+
+   rename "$filename.tmp", $filename or die
+      "rename $filename.tmp -> $filename: $!";
+
+   $lock->close;
+}
+
+=back
+
+=cut
+
 ##################################################################
 #
 # HTML handling code
@@ -753,7 +708,7 @@ BEGIN
     'para'   => "\266", 'middot' => "\267",
     'cedil'  => "\270", 'supl'   => "\271",
     'ordm'   => "\272", 'raquo'  => "\273",
-    'frac14' => "\274", 'frac12 '=> "\275",
+    'frac14' => "\274", 'frac12' => "\275",
     'frac34' => "\276", 'iquest' => "\277",
 
     'Agrave' => "\300", 'Aacute' => "\301",
@@ -1165,7 +1120,7 @@ sub cleanup_cdata {
   ]gesx;
 
   # substitute newlines in the input for html line breaks if required.
-  s%\cM\n%<br />\n%g if $convert_nl;
+  s%\cM?\n%<br />\n%g if $convert_nl;
 
   return $_;
 }
@@ -1175,6 +1130,7 @@ sub cleanup_cdata {
 
 sub escape_html {
   my $str = shift;
+  defined $str or $str = '';
   $str =~ s/([^\w\Q$html_safe_chars\E])/$escape_html_map{$1}/og;
   return $str;
 }
@@ -1194,7 +1150,7 @@ sub unescape_html {
        "&$1$4"
      /gex;
 
-  return $str;
+  return strip_nonprintable($str);
 }
 
 #
